<?php

namespace loyalsoft;

/**
 * 错误级别
 */
class enum_LogLevel extends Enum {

    const All = 0;
    const Info = 1;
    const Warn = 2;
    const Err = 4;

}

/**
 * 日志工具类_高性能(High Performance)
 * @author gwang (mail@wanggangzero.cn)
 * @version <br/>
 *          3.0.2 2018年3月9日 完善稳定性, GetDir()方法添加了自动检查目录是否存在的逻辑. -gwang <br/>
 *          3.0.1 2018年1月9日 整理, online情况下paylog->mysql, errlog->redis, log和info信息仍然在文件(online默认关闭log和info的落地操作). -gwang <br/>
 *          2.0.1 2017年8月8日 将paylog和giftlog和output也保留了.以前向兼容就代码. -gwang <br/>
 *          2.0   2017年8月8日 整理, 将redis部分暂时去掉, 保留缓存模式, 在处理的最后集中flush到文件中. -gwang <br/>
 *          1.0   2016年7月1日 创建, 思路: 建立高性能的日志模块,消除写文件.暂时不包含支付日志. -gwang <br/>
 */
class CLog {

    /**
     * 日志缓存
     * @var array
     */
    private static $Warn = array();
    private static $Info = array();
    private static $Err = array();
    private static $logLevel = enum_LogLevel::Err;

    /**
     * @var int 日志写入redis时保存的记录条数
     */
    const redis_log_trim_max = 5000;

    /**
     * @var int 日志写入redis时每次trim的记录条数
     */
    const redis_log_trim_once = 100;

    /**
     * @return string 日志目录
     */
    private static function GetDir() {
        // 线上版的时候会是linux系统, 单独指定目录, 测试的时候就放到代码旁边得了
        static $dir = null;
        if ($dir) {
            return $dir;
        }
        $dir = (GAME_ONLINE ? "/data/" : ROOTDIR . "/../" ) . "logs/" . PROJECTNAME . "/";
        if (is_dir($dir) || mkdir($dir, 0777, true)) {
            return $dir;
        }
        exit("can not access log directory!");
    }

    /**
     * 设置日志级别
     * @param int $level 参考常量enum_LogLevel :: 0 => info, 1 => log, 2 => err
     */
    public static function SetLogLevel($level) {
        if ($level >= enum_LogLevel::All && $level <= enum_LogLevel::Err) {
            self::$logLevel = $level;
        } else {
            // 无效值,忽视
        }
    }

    /**
     * 断言,
     * @param bool $condition 条件/条件表达式
     * @param string $errmsg 条件不成立时输出到日志中的信息, 日志级别err
     * @param string $okmsg  【可选】默认(null), 当条件成立时不记录任何消息, 除非特别指定了消息内容.日志级别info
     */
    public static function Assert($tag, $condition, $errmsg, $okmsg = null) {
        if (!$condition) {
            self::err($errmsg, $tag);
        } else if ($okmsg) {
            self::info($okmsg, $tag);
        }
    }

    /**
     * 记录日志
     * @param string $msg
     * @param string $tag 标签
     */
    public static function warn($msg, $tag = 'log') {
        if (self::$logLevel <= enum_LogLevel::Warn) {
            self::$Warn[] = self::makeLogMsg($msg, $tag);
        }
    }

    /**
     * @param string $msg
     * @param string $tag 标签
     */
    public static function info($msg, $tag = 'info') {
        if (self::$logLevel <= enum_LogLevel::Info) {
            self::$Info[] = self::makeLogMsg($msg, $tag);
        }
    }

    /**
     * 记录异常日志
     * @param string $msg
     * @param string $tag 标签
     */
    public static function err($msg, $tag = "err") {
        if (self::$logLevel <= enum_LogLevel::Err) {
            self::$Err[] = self::makeLogMsg($msg, $tag);
        }
    }

    /**
     * 将缓存中的数据写入存储设备并清除缓存
     */
    public static function flush() {
        self::flush2File(enum_LogLevel::Info);                                  # 普通信息写到本地文件
        self::flush2File(enum_LogLevel::Warn);
        if (GAME_ONLINE) {
            self::flush2Redis(enum_LogLevel::Err);                              # 外网错误日志写入redis
        } else {
            self::flush2File(enum_LogLevel::Err);                               # 内网错误日志写入本地文件
        }
    }

    /**
     * 将日志写入文件
     * @global type $zoneid
     * @param enum_LogLevel $type
     */
    private static function flush2File($type) {
        $typename = strval(new enum_LogLevel($type));
        $fileName = self::GetDir() . $typename . "-" . date('Ymd') . ".log";    # 日志文件名(按天分割)
        $fd = fopen($fileName, "a");
        if (false === $fd) {                                                    # 打开文件失败
            throw new \Exception("打开 $fileName 失败");
        }
        $arr = self::$$typename;                                                # 取对应类型的数组
        if (count($arr) > 0) {
            foreach ($arr as $msg) {
                $n = fputs($fd, $msg . PHP_EOL);
                if (false === $n) {                                             # 写入时失败
                    throw new \Exception("写入 $fileName 时失败");
                }
                // todo: 其实写入日志也是有失败几率的, 当写入日志也失败的时候,
                // 抛出异常, 让nginx来记录下些东西, 但是如果是在errorhandler
                // 中写日志的时候出错呢......
                //                              -wg 2017年8月5日 08:56:29
            }
            self::$$typename = array();                                         # 清空数据
        }
        fclose($fd);                                                            # 关闭文件
    }

    /**
     * 将缓存中的数据写入存储设备(redis)并清除缓存
     * @type int/enum_LogLevel 日志类型
     */
    private static function flush2Redis($type) {
        $typename = strval(new enum_LogLevel($type));
        $key = "log-" . $typename;
        $arr = self::$$typename;                                                # 取对应类型的数组
        if (count($arr) > 0) {
            redis()->lpush($key, $arr);
            if (redis()->llen($key) > self::redis_log_trim_max + self::redis_log_trim_once) {# 达到清理条件
                redis()->ltrim($key, 0, -self::redis_log_trim_once);            # 缩减记录
            }
            self::$$typename = array();                                         # 清空数据
        }
    }

//
// 避免异常导致缓存丢失, 付费模块的日志直接写入文件
// <editor-fold defaultstate="collapsed" desc=" 付费相关日志, 无缓存 ">

    /**
     * 输出日志 到pay.log
     * @param mixed $msg 参数是字符串则直接输出,否则调用json_encode
     */
    public static function pay($msg) {
        if (GAME_ONLINE) {
            self::Paylog2Mysql($msg);
        } else {
            $fileName = self::GetDir() . "pay.log";
            self::Log2File($fileName, $msg);
        }
    }

    /**
     * 输出日志 到gift.log
     * @param mixed $msg 参数是字符串则直接输出,否则调用json_encode
     * @deprecated since version 0.1
     */
    public static function giftlog($msg) {
        $fileName = self::GetDir() . "gift.log";
        self::Log2File($fileName, $msg);
    }

    /**
     * 向文件写入日志
     * @global int $zoneid
     * @param string $fileName
     * @param string $msg
     */
    private static function Log2File($fileName, $msg) {

        $fd = fopen($fileName, "a");
        if ($fd === FALSE) {                                         # 打开文件失败
            throw new \Exception('open file failed! ' . $fileName);
        }
        // todo: 文件写入也是有机会出错的, 此处出错应予监控,
        // 并抛出异常到nginx/phpfpm层,记录日志.
        //                                -wg 2017年8月5日 09:10:16
        $str = self::makeLogMsg($msg);
        fputs($fd, $str . PHP_EOL);
        fclose($fd);                                                 # 关闭文件
    }

    /**
     * 生成日志,格式: [x区][时间][文件][tag]: msg
     * @global \loyalsoft\type $zoneid
     * @param type $msg
     * @param type $tag
     * @return string
     */
    private static function makeLogMsg($msg, $tag = null) {
        global $zoneid;
        $array = debug_backtrace();
        while (isset($array[0]) && isset($array[0]['file']) && $array[0]['file'] == __FILE__) {
            array_shift($array);
        }
        $row = $array[0];
        $server_addr = isset($_SERVER['SERVER_ADDR']) ? $_SERVER['SERVER_ADDR'] : "-";
        $client_addr = isset($_SERVER['REMOTE_ADDR']) ? $_SERVER['REMOTE_ADDR'] : "-";
        return "[" . $server_addr . "][$zoneid 区][" . date('Y-m-d H:i:s') #              # 主机地址、分区、时间
                . "][" . $client_addr . "][" #                                            # 客户端地址
                . (isset($row['file']) ? CommUtil::str2UTF8(basename($row['file'])) : '-') # 所在文件名
                . ":" . (isset($row['line']) ? $row['line'] : "???")#                     # 所在行   
                . "]" . (isset($tag) ? "[$tag]" : "") . "=> " . var_export($msg, true)#   # 标签、内容
                . PHP_EOL;

//                . "req: " . JsonUtil::encode($req);
    }

    /**
     * 将pay日志写入mysql
     * @param type $msg
     */
    public static function Paylog2Mysql($msg) {
        $str = self::makeLogMsg($msg);                                          # 统一日志格式字符串
        $t = "tab_paylog_" . date('Ym');
        $sql = "create table if not exists $t like tpl_paylog_tab_ym;"
                . " insert into $t (msg) values ('$str');";
        $n = daoInst()->exec($sql);
        if ($n < 0) {
            self::err("写入数据库失败." . $sql);
        }
    }

// </editor-fold>
//
// <editor-fold defaultstate="collapsed" desc=" 待实现">

    /**
     *  将redis中昨天的日志数据输出到磁盘存储介质比如MySQL中.
     *  配合crontable 定时将redis中的日志转存到成本较低的MySQL中方便查阅记录.
     *  而在MySQL中的数据则进行定期备份,比如一个月之后将上个月的数据转储为sql文件,然后从数据库中清除.
     */
    public static function dump() {
        throw new \Exception('uncompleted');
        $redis = gMem();
        $tsday = TimeUtil::dtYesterday();  // 20160702
        $db = CPayInit();
        $sk = "game-run-log-" . $tsday;
        set_time_limit(0);  // 这里要取消超时限制
        self::_insert2mysql($redis, $db, $sk . "-info");
        self::_insert2mysql($redis, $db, $sk . "-log");
        self::_insert2mysql($redis, $db, $sk . "-err");
        $db->close();
    }

    /**
     * 插入记录
     * @param CRedisUtil $redis
     * @param CDBUtil $db
     * @param string $key
     */
    private static function _insert2mysql($redis, $db, $key) {
        $len = 1000; # 一个批次插入1000条记录能达到相对较优的效率了. (100M内网/1000M本机测试,单条记录长度约为200个字节)
        $arr = $redis->lrange($key, 0, $len - 1);
        while (count($arr) > 0) {
            $values = array();
            $sql = 'insert into log201607 (type,ts,tag,msg) values ';
            foreach ($arr as $s) {
                $values[] = "(" . enum_LogLevel::Err . ", $s)";
            }
            $sql .= implode(',', $values);
            $db->query($sql);
            $redis->ltrim($key, $len, -1);
            $arr = $redis->lrange($key, 0, $len - 1);
        }
    }

// </editor-fold>
//
}

if (!GAME_ONLINE) {                                                             # 除了现网
    CLog::SetLogLevel(enum_LogLevel::All);                                      # 开启所有级别的日志输出
}